import * as React from 'react';
export const useAsync = asyncCallback => {
  // Creates a stable callback that manages our loading/success/error status updates
  // as the callback is invoked.
  const storedCallback = React.useRef(asyncCallback);
  storedCallback.current = asyncCallback;

  const createCallback = () => {
    // This callback will not dispatch state once its been cancelled
    let cancelled = false;
    const callback = Object.assign(async (...args) => {
      dispatch({
        status: 'loading'
      });

      try {
        const value = await storedCallback.current(...args);
        !cancelled && dispatch({
          status: 'success',
          value
        });
      } catch (error) {
        !cancelled && dispatch({
          status: 'error',
          error
        });
      }
    }, {
      cancel: () => cancelled = true
    });
    return callback;
  };

  const [state, dispatch] = React.useReducer((prev, action) => ({
    // This is the current status of the promise or async/await function. A
    // promise or async/await can only be in one state at a time.
    status: action.status,
    // The value is persisted between 'success' statuses. This means I can
    // still display things that depend on my current value while my new
    // value is loading.
    value: action.status === 'success' ? action.value : prev.value,
    // Errors get reset each time we leave the error state. There's really
    // no use in keeping those around. They go stale once we leave.
    error: action.status === 'error' ? action.error : void 0,
    callback: action.status === 'cancelled' ? createCallback() : prev.callback
  }), void 0, () => ({
    status: 'idle',
    value: void 0,
    error: void 0,
    callback: createCallback()
  })); // Cancels any pending async callbacks when the hook unmounts

  React.useEffect(() => state.callback.cancel, [state.callback]);

  function _ref() {
    // Prevent the callback from dispatching
    state.callback.cancel(); // Create a new callback and set status to cancelled

    dispatch({
      status: 'cancelled'
    });
  }

  return [React.useMemo(() => ({
    status: state.status,
    value: state.value,
    error: state.error,
    cancel: _ref
  }), [state]), state.callback];
};
export const useAsyncEffect = (asyncCallback, dependencies) => {
  const [state, callback] = useAsync(asyncCallback); // Cancels the current callback and creates a new callback each time the
  // dependencies change

  React.useEffect(() => state.cancel, // eslint-disable-next-line react-hooks/exhaustive-deps
  dependencies); // Runs the callback each time we receive a new one

  React.useEffect(() => {
    callback();
  }, [callback]);
  return state;
};